from datetime import datetime
from . import settings, signature
import random
import time
import re


class SecureAcceptanceAction(object):
    access_key = settings.ACCESS
    currency = settings.DEFAULT_CURRENCY
    date_format = settings.DATE_FORMAT
    locale = settings.LOCALE
    profile_id = settings.PROFILE
    transaction_type = ''


    @property
    def signed_field_names(self):
        return set([
            'access_key',
            'profile_id',
            'transaction_uuid',
            'signed_field_names',
            'unsigned_field_names',
            'signed_date_time',
            'locale',
            'transaction_type',
        ])


    @property
    def unsigned_field_names(self):
        return set([])


    def fields(self):
        names = self.signed_field_names | self.unsigned_field_names
        fields = { name: '' for name in names }

        data, signed_fields = self.build_request_data()
        fields.update(data)

        signed_fields = signed_fields | set(['signed_date_time', 'signed_field_names', 'unsigned_field_names'])
        unsigned_fields = set(fields.keys()) - signed_fields
        fields['signed_date_time'] = datetime.utcnow().strftime(self.date_format)
        fields['signed_field_names'] = ','.join(signed_fields)
        fields['unsigned_field_names'] = ','.join(unsigned_fields)

        signer = signature.SecureAcceptanceSigner()
        fields['signature'] = signer.sign(fields, signed_fields)
        return fields

    def build_request_data(self):
        data = {
            'access_key': self.access_key,
            'currency': self.currency,
            'locale': self.locale,
            'profile_id': self.profile_id,
            'transaction_type': self.transaction_type,
            'transaction_uuid': self.generate_uuid(),
        }

        data.update( self.build_signed_data() )
        signed_fields = self.signed_field_names | set(data.keys())
        data.update( self.build_unsigned_data() )
        return data, signed_fields


    def build_signed_data(self):
        return {}

    def build_unsigned_data(self):
        return {}

    def generate_uuid(self):
        return '%s%s' % (int(time.time()), random.randrange(0, 100))



class ShippingAddressMixin(object):
    def _get_shipping_signed_fields(self):
        return set([
            'ship_to_forename',
            'ship_to_surname',
            'ship_to_address_line1',
            'ship_to_address_line2',
            'ship_to_address_city',
            'ship_to_address_state',
            'ship_to_address_postal_code',
            'ship_to_address_country',
            'ship_to_phone'
        ])

    def _get_shipping_unsigned_fields(self):
        return set([])

    def _get_shipping_data(self):
        if not self.order.shipping_address:
            return {}
        return {
            'ship_to_forename': self.order.shipping_address.first_name,
            'ship_to_surname': self.order.shipping_address.last_name,
            'ship_to_address_line1': self.order.shipping_address.line1,
            'ship_to_address_line2': self.order.shipping_address.line2,
            'ship_to_address_city': self.order.shipping_address.line4,
            'ship_to_address_state': self.order.shipping_address.state,
            'ship_to_address_postal_code': self.order.shipping_address.postcode,
            'ship_to_address_country': self.order.shipping_address.country.code,
            'ship_to_phone': re.sub('[^0-9]', '', self.order.shipping_address.phone_number.as_rfc3966),
        }



class BillingAddressMixin(object):
    def _get_billing_signed_fields(self):
        return set([
            'bill_to_forename',
            'bill_to_surname',
            'bill_to_address_line1',
            'bill_to_address_line2',
            'bill_to_address_city',
            'bill_to_address_state',
            'bill_to_address_postal_code',
            'bill_to_address_country',
            'bill_to_email',
        ])

    def _get_billing_unsigned_fields(self):
        return set([
            # Oscar doesn't track phone number for billing addresses. Set this as unsigned to that the client JS can specify it if they want.
            'bill_to_phone',
        ])

    def _get_billing_data(self):
        data = {
            'bill_to_email': self.order.email,
        }
        if self.order.billing_address:
            data['bill_to_forename'] = self.order.billing_address.first_name
            data['bill_to_surname'] = self.order.billing_address.last_name
            data['bill_to_address_line1'] = self.order.billing_address.line1
            data['bill_to_address_line2'] = self.order.billing_address.line2
            data['bill_to_address_city'] = self.order.billing_address.line4
            data['bill_to_address_state'] = self.order.billing_address.state
            data['bill_to_address_postal_code'] = self.order.billing_address.postcode
            data['bill_to_address_country'] = self.order.billing_address.country.code
        return data


class OrderAction(SecureAcceptanceAction, ShippingAddressMixin, BillingAddressMixin):
    """
    Abstract SecureAcceptanceAction for action's related to orders.
    """
    def __init__(self, order, amount, **kwargs):
        self.order = order
        self.amount = amount
        self.customer_ip_address = kwargs.get('customer_ip_address')
        self.device_fingerprint_id = kwargs.get('fingerprint_session_id')
        self.extra_fields = kwargs.get('extra_fields')


    @property
    def signed_field_names(self):
        fields = super().signed_field_names
        fields = fields | self._get_shipping_signed_fields()
        fields = fields | self._get_billing_signed_fields()
        fields = fields | set([
            'payment_method',
            'reference_number',
            'currency',
            'amount',
            'line_item_count',
            'customer_ip_address',
            'device_fingerprint_id',
        ])
        return fields


    @property
    def unsigned_field_names(self):
        fields = super().unsigned_field_names
        fields = fields | self._get_shipping_unsigned_fields()
        fields = fields | self._get_billing_unsigned_fields()
        return fields


    def build_signed_data(self):
        data = {}

        # Basic order info
        data['payment_method'] = 'card'
        data['reference_number'] = str(self.order.number)
        data['currency'] = self.order.currency
        data['amount'] = str(self.amount)

        # Add shipping and billing info
        data.update(self._get_shipping_data())
        data.update(self._get_billing_data())

        # Add line item info
        i = 0
        for line in self.order.lines.all():
            data['item_%s_name' % i] = line.product.title
            data['item_%s_sku' % i] = line.partner_sku
            data['item_%s_quantity' % i] = str(line.quantity)
            data['item_%s_unit_price' % i] = str(line.unit_price_incl_tax)
            i += 1
        data['line_item_count'] = str(i)

        # Other misc fields
        for field in ('customer_ip_address', 'device_fingerprint_id'):
            value = getattr(self, field, None)
            if value is not None:
                data[field] = value
        data.update(self.extra_fields)

        return data


class CreatePaymentToken(OrderAction):
    transaction_type = 'create_payment_token'
    url = settings.ENDPOINT_PAY

    @property
    def unsigned_field_names(self):
        fields = super().unsigned_field_names
        fields = fields | set([
            # There are unsigned as the server should *never* know them. The client JS must fill them in.
            'card_type',
            'card_number',
            'card_expiry_date',
            'card_cvn',
        ])
        return fields


class AuthorizePaymentToken(OrderAction):
    transaction_type = 'authorization'
    url = settings.ENDPOINT_PAY

    def __init__(self, token_string, order, amount, **kwargs):
        self.token_string = token_string
        super().__init__(order, amount, **kwargs)


    @property
    def signed_field_names(self):
        fields = super().signed_field_names
        fields = fields | set([
            'payment_token',
        ])
        return fields


    def build_signed_data(self):
        data = super().build_signed_data()
        data['payment_token'] = self.token_string
        return data



from django.core.exceptions import SuspiciousOperation
from . import settings
import hashlib
import hmac
import base64


class SecureAcceptanceSigner(object):
    secret_key = settings.SECRET

    def sign(self, data, signed_fields):
        key = self.secret_key.encode('utf-8')
        msg_raw = self._build_message(data, signed_fields).encode('utf-8')
        msg_hmac = hmac.new(key, msg_raw, hashlib.sha256)
        return base64.b64encode(msg_hmac.digest())

    def verify_request(self, request):
        # Ensure the signature is valid and that this request can be trusted
        signed_field_names = request.POST.get('signed_field_names')
        if not signed_field_names:
            raise SuspiciousOperation("Request has no fields to verify")
        signed_field_names = signed_field_names.split(',')
        signature_given = request.POST['signature'].encode('utf-8')
        signature_calc = self.sign(request.POST, signed_field_names)
        return (signature_given == signature_calc)

    def _build_message(self, data, signed_fields):
        parts = []
        for field in signed_fields:
            parts.append( '%s=%s' % (field, data.get(field, '')) )
        return ','.join(parts)



from django.test import TestCase
from django.test.client import RequestFactory
from ..signature import SecureAcceptanceSigner


class SignerTest(TestCase):
    def test_sign(self):
        signer = SecureAcceptanceSigner()

        # Baseline
        signer.secret_key = 'FOO'
        signature = signer.sign({ 'foo': 'bar', 'baz': 'bat' }, ('foo', 'baz'))
        self.assertEqual(signature, b'IVMC7Aj8pDKwLx+0eNfIfoQAHvViiLeavLyYatCtB+c=')

        # Change field order
        signature = signer.sign({ 'foo': 'bar', 'baz': 'bat' }, ('baz', 'foo'))
        self.assertEqual(signature, b'5Gw1ffUlVU9Cm0tTa/nzhQ81Bc6/SDqz/tEP5VyMzkk=')

        # Change key
        signer.secret_key = 'SECRET'
        signature = signer.sign({ 'foo': 'bar', 'baz': 'bat' }, ('foo', 'baz'))
        self.assertEqual(signature, b'FvjC1PIhxuaLipTbRDw9UXL6F58t9Hyj12HLHiYoOD0=')

    def test_verify(self):
        rf = RequestFactory()
        signer = SecureAcceptanceSigner()

        # Baseline
        signer.secret_key = 'FOO'
        request = rf.post('/', {
            'signed_field_names': 'foo,baz',
            'signature': 'IVMC7Aj8pDKwLx+0eNfIfoQAHvViiLeavLyYatCtB+c=',
            'foo': 'bar',
            'baz': 'bat',
        })
        self.assertTrue( signer.verify_request(request) )

        # Bad signature given
        request = rf.post('/', {
            'signed_field_names': 'foo,baz',
            'signature': 'IVMC7Aj8pDKwLx+0eNfIfoQAHvViiLeavLyYatCtB',
            'foo': 'bar',
            'baz': 'bat',
        })
        self.assertFalse( signer.verify_request(request) )



from datetime import datetime
from django.db import models
from django.contrib.postgres.fields import HStoreField
from oscar.core.compat import AUTH_USER_MODEL


class CyberSourceReply(models.Model):
    user = models.ForeignKey(AUTH_USER_MODEL,
        related_name='cybersource_replies', null=True, blank=True, on_delete=models.SET_NULL)
    data = HStoreField()
    date_modified = models.DateTimeField("Date Modified", auto_now=True)
    date_created = models.DateTimeField("Date Received", auto_now_add=True)

    def __str__(self):
        return 'CyberSource Reply %s' % self.date_created


class ReplyLogMixin(object):
    def log_field(self, key, default=''):
        return self.log.data.get(key, default)


class PaymentToken(ReplyLogMixin, models.Model):
    TYPES = {
        '001': 'Visa',
        '002': 'MasterCard',
        '003': 'Amex',
        '004': 'Discover',
        '005': 'DINERS_CLUB',
        '006': 'CARTE_BLANCHE',
        '007': 'JCB',
        '014': 'ENROUTE',
        '021': 'JAL',
        '024': 'MAESTRO_UK_DOMESTIC',
        '031': 'DELTA',
        '033': 'VISA_ELECTRON',
        '034': 'DANKORT',
        '036': 'CARTE_BLEUE',
        '037': 'CARTA_SI',
        '042': 'MAESTRO_INTERNATIONAL',
        '043': 'GE_MONEY_UK_CARD',
        '050': 'HIPERCARD',
        '054': 'ELO',
    }

    log = models.ForeignKey(CyberSourceReply, related_name='tokens')
    token = models.CharField(max_length=100, unique=True)
    masked_card_number = models.CharField(max_length=25)
    card_type = models.CharField(max_length=10)

    @property
    def card_type_name(self):
        return self.TYPES.get(self.card_type)

    @property
    def billing_zip_code(self):
        return self.log_field('req_bill_to_address_postal_code')

    @property
    def expiry_month(self):
        return self.log_field('req_card_expiry_date').split('-')[0]

    @property
    def expiry_year(self):
        return self.log_field('req_card_expiry_date').split('-')[1]

    @property
    def card_last4(self):
        return self.masked_card_number[-4:]

    @property
    def card_holder(self):
        return "%s %s" % (self.log_field('req_bill_to_forename'), self.log_field('req_bill_to_surname'))

    def __str__(self):
        return "%s" % self.masked_card_number


class TransactionMixin(ReplyLogMixin, models.Model):
    log = models.ForeignKey(CyberSourceReply,
        related_name='transactions',
        null=True,
        blank=True,
        on_delete=models.SET_NULL)
    token = models.ForeignKey(PaymentToken,
        related_name='transactions',
        null=True,
        blank=True,
        on_delete=models.SET_NULL)
    request_token = models.CharField(max_length=200, null=True, blank=True)
    processed_datetime = models.DateTimeField(default=datetime.now)

    class Meta:
        abstract = True



import django.dispatch

pre_build_get_token_request = django.dispatch.Signal(providing_args=["extra_fields", "request", "order", "source"])
pre_build_auth_request = django.dispatch.Signal(providing_args=["extra_fields", "request", "order", "token"])



from django.contrib import admin
from . import models


@admin.register(models.PaymentToken)
class PaymentTokenAdmin(admin.ModelAdmin):
    list_filter = ['card_type', 'log__date_created']
    search_fields = ['token', 'card_type', 'masked_card_number']
    fields = ['token', 'card_type', 'masked_card_number', 'log']
    list_display = ['token', 'card_type', 'masked_card_number', 'log']
    readonly_fields = fields


@admin.register(models.CyberSourceReply)
class CyberSourceReplyAdmin(admin.ModelAdmin):
    list_filter = ['date_modified', 'date_created']
    list_display = ['date_created', 'user', 'date_modified']
    fields = ['user', 'data', 'date_modified', 'date_created']
    readonly_fields = fields



from datetime import datetime
from decimal import Decimal
from oscar.core.loading import get_model
from oscarapicheckout.methods import PaymentMethod, PaymentMethodSerializer
from oscarapicheckout.states import FormPostRequired, Complete, Declined
from .constants import CHECKOUT_FINGERPRINT_SESSION_ID
from .models import PaymentToken
from . import actions, signals, settings

Transaction = get_model('payment', 'Transaction')


class Cybersource(PaymentMethod):
    """
    This is an example of how to implement a payment method that required some off-site
    interaction, like Cybersource Secure Acceptance, for example. It returns a pending
    status initially that requires the client app to make a form post, which in-turn
    redirects back to us. This is a common pattern in PCI SAQ A-EP ecommerce sites.
    """
    name = 'Cybersource Secure Acceptance'
    code = 'cybersource'
    serializer_class = PaymentMethodSerializer


    # Payment Step 1: Require form POST to Cybersource
    def _record_payment(self, request, order, amount, reference, **kwargs):
        source = self.get_source(order, reference)
        amount_to_allocate = amount - source.amount_allocated

        # Allow application to include extra, arbitrary fields in the request to CS
        extra_fields = {}
        signals.pre_build_get_token_request.send(
            sender=self.__class__,
            extra_fields=extra_fields,
            request=request,
            order=order,
            source=source)

        # Build the data for CyberSource transaction
        operation = actions.CreatePaymentToken(
            order=order,
            amount=amount_to_allocate,
            customer_ip_address=request.META['REMOTE_ADDR'],
            fingerprint_session_id=request.session.get(CHECKOUT_FINGERPRINT_SESSION_ID),
            extra_fields=extra_fields)

        # Return form fields to the browser. The browser then needs to fill in the blank
        # fields (like billing data) and submit them as a POST to CyberSource.
        url, fields = self._fields(operation)

        # Return a response showing we need to post some fields to the given
        # URL to finishing processing this payment method
        return FormPostRequired(
            amount=amount,
            name='get-token',
            url=url,
            fields=fields)


    # Payment Step 2: Record the generated payment token and require authorization using the token.
    def record_created_payment_token(self, request, reply_log_entry, order, data):
        token_string = data.get('payment_token')
        card_num = data.get('req_card_number')
        card_type = data.get('req_card_type')
        req_amount = Decimal(data.get('req_amount', '0'))

        # Create the payment token
        try:
            token = PaymentToken.objects.get(token=token_string)
        except PaymentToken.DoesNotExist:
            token = PaymentToken(
                log=reply_log_entry,
                token=token_string,
                masked_card_number=card_num,
                card_type=card_type)
            token.save()

        # Allow application to include extra, arbitrary fields in the request to CS
        extra_fields = {}
        signals.pre_build_auth_request.send(
            sender=self.__class__,
            extra_fields=extra_fields,
            order=order,
            token=token)

        # Build the data for CyberSource transaction
        operation = actions.AuthorizePaymentToken(
            token_string=token_string,
            order=order,
            amount=req_amount,
            customer_ip_address=request.META['REMOTE_ADDR'],
            fingerprint_session_id=request.session.get(CHECKOUT_FINGERPRINT_SESSION_ID),
            extra_fields=extra_fields)

        # Return form fields to the browser. The browser then needs to fill in the blank
        # fields (like billing data) and submit them as a POST to CyberSource.
        url, fields = self._fields(operation)

        # Require that the client now perform an authorization using this token
        return FormPostRequired(
            amount=req_amount,
            name='authorize',
            url=url,
            fields=fields)


    # Payment Step 3: Record a successful authorization
    def record_successful_authorization(self, reply_log_entry, order, data):
        token_string = data.get('req_payment_token')
        transaction_id = data.get('transaction_id')
        decision = data.get('decision')
        request_token = data.get('request_token')
        signed_date_time = data.get('signed_date_time')
        auth_amount = Decimal(data.get('auth_amount', '0'))
        req_amount = Decimal(data.get('req_amount', '0'))

        try:
            token = PaymentToken.objects.get(token=token_string)
        except PaymentToken.DoesNotExist:
            return Declined(req_amount)

        source = self.get_source(order, transaction_id)
        source.amount_allocated += auth_amount
        source.save()

        transaction = Transaction()
        transaction.log = reply_log_entry
        transaction.source = source
        transaction.token = token
        transaction.txn_type = Transaction.AUTHORISE
        transaction.amount = req_amount
        transaction.reference = transaction_id
        transaction.status = decision
        transaction.request_token = request_token
        transaction.processed_datetime = datetime.strptime(signed_date_time, settings.DATE_FORMAT)
        transaction.save()

        event = self.make_authorize_event(order, auth_amount, transaction_id)
        for line in order.lines.all():
            self.make_event_quantity(event, line, line.quantity)

        return Complete(source.amount_allocated)


    # Payment Declined in Step 3: Record a failed authorization.
    def record_declined_authorization(self, reply_log_entry, order, data):
        token_string = data.get('req_payment_token')
        transaction_id = data.get('transaction_id')
        decision = data.get('decision')
        request_token = data.get('request_token')
        signed_date_time = data.get('signed_date_time')
        req_amount = Decimal(data.get('req_amount', '0'))

        transaction = Transaction()
        transaction.log = reply_log_entry
        transaction.source = self.get_source(order, transaction_id)
        transaction.token = PaymentToken.objects.first(token=token_string)
        transaction.txn_type = Transaction.AUTHORISE
        transaction.amount = req_amount
        transaction.reference = transaction_id
        transaction.status = decision
        transaction.request_token = request_token
        transaction.processed_datetime = datetime.strptime(signed_date_time, settings.DATE_FORMAT)
        transaction.save()

        return Declined(req_amount)


    def _fields(self, operation):
        """Helper to convert an operation object into a list of fields and a URL"""
        fields = []
        cs_fields = operation.fields()
        editable_fields = cs_fields['unsigned_field_names'].split(',')
        for key, value in cs_fields.items():
            fields.append({
                'key': key,
                'value': value if isinstance(value, str) else value.decode(),
                'editable': (key in editable_fields)
            })

        return operation.url, fields



from django.test import TestCase

from ..models import CyberSourceReply, PaymentToken
from .factories import build_accepted_token_reply_data


class PaymentTokenTest(TestCase):
    def test_log_data_parsing(self):
        data = build_accepted_token_reply_data('S123456789')
        log = CyberSourceReply.objects.create(data=data)
        token = PaymentToken.objects.create(
            log=log,
            token=data['payment_token'],
            masked_card_number=data['req_card_number'],
            card_type=data['req_card_type'])

        self.assertEquals(token.card_type_name, 'Visa')
        self.assertEquals(token.billing_zip_code, '10001')
        self.assertEquals(token.expiry_month, '12')
        self.assertEquals(token.expiry_year, '2020')
        self.assertEquals(token.card_last4, '1111')
        self.assertEquals(token.card_holder, 'Bob Smith')



from datetime import datetime
from random import randrange
from ..signature import SecureAcceptanceSigner
import uuid


def build_accepted_token_reply_data(order_number):
    data = {
        "decision": "ACCEPT",
        "payment_token": "4600379961546299901519",
        "reason_code": "100",
        "req_access_key": "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
        "req_amount": "99.99",
        "req_bill_to_address_city": "Manhattan",
        "req_bill_to_address_country": "US",
        "req_bill_to_address_line1": "627 W 27th st",
        "req_bill_to_address_postal_code": "10001",
        "req_bill_to_address_state": "NY",
        "req_bill_to_email": "foo@example.com",
        "req_bill_to_forename": "Bob",
        "req_bill_to_phone": "18006927753",
        "req_bill_to_surname": "Smith",
        "req_card_expiry_date": "12-2020",
        "req_card_number": "xxxxxxxxxxxx1111",
        "req_card_type": "001",
        "req_currency": "USD",
        "req_customer_ip_address": "8.8.8.8",
        "req_device_fingerprint_id": str(uuid.uuid4()),
        "req_item_0_name": "My Product",
        "req_item_0_quantity": "1",
        "req_item_0_sku": "123456789",
        "req_item_0_unit_price": "99.99",
        "req_line_item_count": "1",
        "req_locale": "en",
        "req_payment_method": "card",
        "req_profile_id": str(uuid.uuid4()),
        "req_reference_number": order_number,
        "req_ship_to_address_city": "Manhattan",
        "req_ship_to_address_country": "US",
        "req_ship_to_address_line1": "627 W 27th st",
        "req_ship_to_address_postal_code": "10001",
        "req_ship_to_address_state": "NY",
        "req_ship_to_forename": "Bob",
        "req_ship_to_phone": "18006927753",
        "req_ship_to_surname": "Smith",
        "req_transaction_type": "create_payment_token",
        "req_transaction_uuid": str(randrange(100000000000, 999999999999)),
        "request_token": "Ahj/7wSR8sYxolZgxwyeIkG7lw3ZNnDmLNjMqcSPOS4lfDoTAFLiV8OhM0glqN4vpQyaSZbpAd0+3AnI+WMY0SswY4ZPAAAALRaM",
        "transaction_id": str(randrange(0, 99999999999999999999)),
        "utf8": "✓",
    }
    return data


def build_accepted_auth_reply_data(order_number):
    data = {
        "auth_amount": "99.99",
        "auth_avs_code": "X",
        "auth_avs_code_raw": "I1",
        "auth_code": "888888",
        "auth_response": "100",
        "auth_time": "2016-04-07T140637Z",
        "auth_trans_ref_no": "79872689EMF2SDGN",
        "decision": "ACCEPT",
        "decision_case_priority": "3",
        "decision_early_rcode": "1",
        "decision_early_reason_code": "100",
        "decision_early_return_code": "9999999",
        "decision_rcode": "1",
        "decision_reason_code": "100",
        "decision_return_code": "1320000",
        "decision_rflag": "SOK",
        "decision_rmsg": "Service processed successfully",
        "message": "Request was processed successfully.",
        "req_payment_token": "4600379961546299901519",
        "reason_code": "100",
        "req_access_key": "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
        "req_amount": "99.99",
        "req_bill_to_address_city": "Manhattan",
        "req_bill_to_address_country": "US",
        "req_bill_to_address_line1": "627 W 27th st",
        "req_bill_to_address_postal_code": "10001",
        "req_bill_to_address_state": "NY",
        "req_bill_to_email": "foo@example.com",
        "req_bill_to_forename": "Bob",
        "req_bill_to_phone": "18006927753",
        "req_bill_to_surname": "Smith",
        "req_card_expiry_date": "12-2020",
        "req_card_number": "xxxxxxxxxxxx1111",
        "req_card_type": "001",
        "req_currency": "USD",
        "req_customer_ip_address": "8.8.8.8",
        "req_device_fingerprint_id": str(uuid.uuid4()),
        "req_item_0_name": "My Product",
        "req_item_0_quantity": "1",
        "req_item_0_sku": "123456789",
        "req_item_0_unit_price": "99.99",
        "req_line_item_count": "1",
        "req_locale": "en",
        "req_payment_method": "card",
        "req_profile_id": str(uuid.uuid4()),
        "req_reference_number": order_number,
        "req_ship_to_address_city": "Manhattan",
        "req_ship_to_address_country": "US",
        "req_ship_to_address_line1": "627 W 27th st",
        "req_ship_to_address_postal_code": "10001",
        "req_ship_to_address_state": "NY",
        "req_ship_to_forename": "Bob",
        "req_ship_to_phone": "18006927753",
        "req_ship_to_surname": "Smith",
        "req_transaction_type": "authorization",
        "req_transaction_uuid": str(randrange(100000000000, 999999999999)),
        "request_token": "Ahj/7wSR8sYxolZgxwyeIkG7lw3ZNnDmLNjMqcSPOS4lfDoTAFLiV8OhM0glqN4vpQyaSZbpAd0+3AnI+WMY0SswY4ZPAAAALRaM",
        "score_bin_country": "US",
        "score_card_issuer": "JPMORGAN CHASE BANK, N.A.",
        "score_card_scheme": "VISA CREDIT",
        "score_host_severity": "1",
        "score_internet_info": "FREE-EM^MM-IPBC",
        "score_ip_city": "new york",
        "score_ip_country": "us",
        "score_ip_routing_method": "fixed",
        "score_ip_state": "ny",
        "score_model_used": "default",
        "score_rcode": "1",
        "score_reason_code": "100",
        "score_return_code": "1070000",
        "score_rflag": "SOK",
        "score_rmsg": "score service was successful",
        "score_score_result": "0",
        "score_time_local": "10:06:36",
        "transaction_id": str(randrange(0, 99999999999999999999)),
        "utf8": "✓",
    }
    return data


def build_declined_token_reply_data(order_number):
    data = {
        "decision": "DECLINE",
        "message": "We encountered a Paymentech problem: Reason: Processor Decline.",
        "reason_code": "203",
        "req_access_key": "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
        "req_amount": "99.99",
        "req_bill_to_address_city": "Manhattan",
        "req_bill_to_address_country": "US",
        "req_bill_to_address_line1": "627 W 27th st",
        "req_bill_to_address_line2": "",
        "req_bill_to_address_postal_code": "10001",
        "req_bill_to_address_state": "NY",
        "req_bill_to_email": "foo@example.com",
        "req_bill_to_forename": "Bob",
        "req_bill_to_phone": "18006927753",
        "req_bill_to_surname": "Smith",
        "req_card_expiry_date": "12-2020",
        "req_card_number": "xxxxxxxxxxxx1111",
        "req_card_type": "001",
        "req_currency": "USD",
        "req_customer_ip_address": "8.8.8.8",
        "req_device_fingerprint_id": str(uuid.uuid4()),
        "req_item_0_name": "My Product",
        "req_item_0_quantity": "1",
        "req_item_0_sku": "12345678",
        "req_item_0_unit_price": "99.99",
        "req_line_item_count": "1",
        "req_locale": "en",
        "req_payment_method": "card",
        "req_profile_id": str(uuid.uuid4()),
        "req_reference_number": order_number,
        "req_ship_to_address_city": "Manhattan",
        "req_ship_to_address_country": "US",
        "req_ship_to_address_line1": "627 W 27th st",
        "req_ship_to_address_line2": "",
        "req_ship_to_address_postal_code": "10001",
        "req_ship_to_address_state": "NY",
        "req_ship_to_forename": "Bob",
        "req_ship_to_phone": "18006927753",
        "req_ship_to_surname": "Smith",
        "req_transaction_type": "create_payment_token",
        "req_transaction_uuid": str(randrange(100000000000, 999999999999)),
        "request_token": "Ahj77wSR8sY0PySenECSIpcaMXXZAClxoxddkaQR1G8vpJl6SZbpAd0+2AnI+WMaH5JPTiBJAAAA9hKE",
        "transaction_id": str(randrange(0, 99999999999999999999)),
        "utf8": "✓",
    }
    return data


def build_declined_auth_reply_data(order_number):
    data = {
        "auth_avs_code": "X",
        "auth_avs_code_raw": "I1",
        "auth_response": "303",
        "decision": "DECLINE",
        "decision_early_rcode": "1",
        "decision_early_reason_code": "100",
        "decision_early_return_code": "9999999",
        "message": "We encountered a Paymentech problem: Reason: Processor Decline.",
        "reason_code": "203",
        "req_access_key": "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
        "req_amount": "99.99",
        "req_bill_to_address_city": "Manhattan",
        "req_bill_to_address_country": "US",
        "req_bill_to_address_line1": "627 W 27th st",
        "req_bill_to_address_line2": "",
        "req_bill_to_address_postal_code": "10001",
        "req_bill_to_address_state": "NY",
        "req_bill_to_email": "foo@example.com",
        "req_bill_to_forename": "Bob",
        "req_bill_to_phone": "18006927753",
        "req_bill_to_surname": "Smith",
        "req_card_expiry_date": "12-2020",
        "req_card_number": "xxxxxxxxxxxx1111",
        "req_card_type": "001",
        "req_currency": "USD",
        "req_customer_ip_address": "8.8.8.8",
        "req_device_fingerprint_id": str(uuid.uuid4()),
        "req_item_0_name": "My Product",
        "req_item_0_quantity": "1",
        "req_item_0_sku": "12345678",
        "req_item_0_unit_price": "99.99",
        "req_line_item_count": "1",
        "req_locale": "en",
        "req_payment_method": "card",
        "req_profile_id": str(uuid.uuid4()),
        "req_reference_number": order_number,
        "req_ship_to_address_city": "Manhattan",
        "req_ship_to_address_country": "US",
        "req_ship_to_address_line1": "627 W 27th st",
        "req_ship_to_address_line2": "",
        "req_ship_to_address_postal_code": "10001",
        "req_ship_to_address_state": "NY",
        "req_ship_to_forename": "Bob",
        "req_ship_to_phone": "18006927753",
        "req_ship_to_surname": "Smith",
        "req_transaction_type": "authorization",
        "req_transaction_uuid": str(randrange(100000000000, 999999999999)),
        "request_token": "Ahj77wSR8sY0PySenECSIpcaMXXZAClxoxddkaQR1G8vpJl6SZbpAd0+2AnI+WMaH5JPTiBJAAAA9hKE",
        "transaction_id": str(randrange(0, 99999999999999999999)),
        "utf8": "✓",
    }
    return data


def sign_reply_data(data):
    fields = list(data.keys())
    data['signature'] = SecureAcceptanceSigner().sign(data, fields).decode('utf8')
    data['signed_date_time'] = datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%SZ')
    data['signed_field_names'] = ','.join(fields)
    return data



from decimal import Decimal
from django.contrib import messages
from django.core.exceptions import SuspiciousOperation
from django.db import transaction
from django.http import Http404
from django.shortcuts import redirect
from django.views import generic
from rest_framework.views import APIView
from oscar.core.loading import get_class, get_model
from oscarapicheckout import utils
from . import actions, settings, signature
from .authentication import CSRFExemptSessionAuthentication
from .constants import CHECKOUT_FINGERPRINT_SESSION_ID, DECISION_ACCEPT
from .methods import Cybersource
from .models import CyberSourceReply
import uuid
import logging

OrderNumberGenerator = get_class('order.utils', 'OrderNumberGenerator')
OrderPlacementMixin = get_class('checkout.mixins', 'OrderPlacementMixin')
OrderTotalCalculator = get_class('checkout.calculators', 'OrderTotalCalculator')

Basket = get_model('basket', 'Basket')
BillingAddress = get_model('order', 'BillingAddress')
Country = get_model('address', 'Country')
Order = get_model('order', 'Order')
PaymentEventType = get_model('order', 'PaymentEventType')
PaymentEvent = get_model('order', 'PaymentEvent')
PaymentEventQuantity = get_model('order', 'PaymentEventQuantity')
ShippingAddress = get_model('order', 'ShippingAddress')
Source = get_model('payment', 'Source')
SourceType = get_model('payment', 'SourceType')
Transaction = get_model('payment', 'Transaction')


logger = logging.getLogger(__name__)



class FingerprintRedirectView(generic.View):
    url_types = {
        'img-1': '%(protocol)s://%(host)s/fp/clear.png?org_id=%(org_id)s&session_id=%(merchant_id)s%(session_id)s&m=1',
        'img-2': '%(protocol)s://%(host)s/fp/clear.png?org_id=%(org_id)s&session_id=%(merchant_id)s%(session_id)s&m=2',
        'flash': '%(protocol)s://%(host)s/fp/fp.swf?org_id=%(org_id)s&session_id=%(merchant_id)s%(session_id)s',
        'js': '%(protocol)s://%(host)s/fp/check.js?org_id=%(org_id)s&session_id=%(merchant_id)s%(session_id)s',
    }

    def get(self, request, url_type):
        if url_type not in self.url_types:
            raise Http404('url_type not found')

        sessid = request.session.get(CHECKOUT_FINGERPRINT_SESSION_ID)
        if not sessid:
            sessid = str(uuid.uuid1())
            request.session[CHECKOUT_FINGERPRINT_SESSION_ID] = sessid

        data = {
            'protocol': settings.FINGERPRINT_PROTOCOL,
            'host': settings.FINGERPRINT_HOST,
            'org_id': settings.ORG_ID,
            'merchant_id': settings.MERCHANT_ID,
            'session_id': sessid,
        }
        url = self.url_types[url_type] % data
        return redirect(url)


class CyberSourceReplyView(APIView):
    """
    Handle a CyberSource reply.
    """
    authentication_classes = (CSRFExemptSessionAuthentication, )


    def post(self, request, format=None):
        if not self.is_request_valid(request):
            raise SuspiciousOperation('Bad Signature')

        # Record in reply log
        log = self.log_response(request)

        # Invoke handler for transaction type
        trans_type = request.data.get('req_transaction_type')
        handler = self.get_handler_fn(trans_type)
        with transaction.atomic():
            resp = handler(request, format, log)
        return resp


    def is_request_valid(self, request):
        return signature.SecureAcceptanceSigner().verify_request(request)


    def log_response(self, request):
        log = CyberSourceReply(
            user=request.user if request.user.is_authenticated() else None,
            data=request.data)
        log.save()
        return log


    def get_handler_fn(self, trans_type):
        handlers = {
            actions.CreatePaymentToken.transaction_type: self.record_token,
            actions.AuthorizePaymentToken.transaction_type: self.record_authorization,
        }
        if trans_type not in handlers:
            raise SuspiciousOperation("Couldn't find handler for %s" % trans_type)
        return handlers[trans_type]


    def record_token(self, request, format, reply_log_entry):
        # Fetch the related order
        order = self._get_order(request)

        # Check if the payment token was actually created or not.
        if request.data.get('decision') != DECISION_ACCEPT:
            messages.add_message(request._request, messages.ERROR, settings.CARD_REJECT_ERROR)
            amount = Decimal(request.data.get('req_amount', '0.00'))
            utils.mark_payment_method_declined(order, request, Cybersource.code, amount)
            return redirect(settings.REDIRECT_FAIL)

        # Token creation was successful, so log it and move on to authorization.
        new_state = Cybersource().record_created_payment_token(request, reply_log_entry, order, request.data)
        utils.update_payment_method_state(order, request, Cybersource.code, new_state)
        return redirect(settings.REDIRECT_PENDING)


    def record_authorization(self, request, format, reply_log_entry):
        # If the transaction already exists, do nothing
        if Transaction.objects.filter(reference=request.data.get('transaction_id')).exists():
            logger.warning('Duplicate transaction_id received from CyberSource: %s' % request.data.get('transaction_id'))
            return redirect(settings.REDIRECT_SUCCESS)

        # Fetch the related order
        order = self._get_order(request)

        # Check if the authorization was declined
        if request.data.get('decision') != DECISION_ACCEPT:
            messages.add_message(request._request, messages.ERROR, settings.CARD_REJECT_ERROR)
            amount = Decimal(request.data.get('req_amount', '0.00'))
            utils.mark_payment_method_declined(order, request, Cybersource.code, amount)
            return redirect(settings.REDIRECT_FAIL)

        # Authorization was successful, so log it and redirect to the success page.
        new_state = Cybersource().record_successful_authorization(reply_log_entry, order, request.data)
        utils.update_payment_method_state(order, request, Cybersource.code, new_state)
        return redirect(settings.REDIRECT_SUCCESS)


    def _get_order(self, request):
        try:
            order = Order.objects.get(number=request.data.get('req_reference_number'))
        except Order.DoesNotExist:
            raise SuspiciousOperation("Order not found.")
        return order



from django.conf.urls import url
from django.views.decorators.csrf import csrf_exempt
from oscar.core.application import Application
from .views import CyberSourceReplyView, FingerprintRedirectView


class CybersourceApplication(Application):
    def get_urls(self):
        cs_reply = csrf_exempt(CyberSourceReplyView.as_view())
        fingerprint = FingerprintRedirectView.as_view()

        urlpatterns = [
            url(r'^cybersource-reply/$', cs_reply, name='cybersource-reply'),
            url(r'^fingerprint/(?P<url_type>.*)/$', fingerprint, name='cybersource-fingerprint-redirect'),
        ]
        return self.post_process_urls(urlpatterns)


application = CybersourceApplication()



from bs4 import BeautifulSoup
from decimal import Decimal as D
from django.core import mail
from django.core.urlresolvers import reverse
from mock import patch
from oscar.core.loading import get_model
from oscar.test import factories
from rest_framework import status
from rest_framework.test import APITestCase
import datetime
import requests  # Needed for external calls

from ..constants import CHECKOUT_ORDER_ID
from . import factories as cs_factories

Basket = get_model('basket', 'Basket')
Product = get_model('catalogue', 'Product')
Order = get_model('order', 'Order')


class BaseCheckoutTest(APITestCase):
    fixtures = ['cybersource-test.yaml']

    def create_product(self, price=D('10.00')):
        product = factories.create_product(
            title='My Product',
            product_class='My Product Class')
        record = factories.create_stockrecord(
            currency='USD',
            product=product,
            num_in_stock=10,
            price_excl_tax=price)
        factories.create_purchase_info(record)
        return product


    def do_add_to_basket(self, product_id, quantity=1):
        url = reverse('api-basket-add-product')
        data = {
            "url": reverse('product-detail', args=[product_id]),
            "quantity": quantity
        }
        return self.client.post(url, data)


    def do_get_basket(self):
        url = reverse('api-basket')
        return self.client.get(url)


    def do_checkout(self, basket_id, extra_data={}):
        data = {
            "guest_email": "joe@example.com",
            "basket": reverse('basket-detail', args=[basket_id]),
            "shipping_address": {
                "first_name": "Joe",
                "last_name": "Schmoe",
                "line1": "234 5th Ave",
                "line4": "Manhattan",
                "postcode": "10001",
                "state": "NY",
                "country": reverse('country-detail', args=['US']),
                "phone_number": "+1 (717) 467-1111",
            },
            "billing_address": {
                "first_name": "Joe",
                "last_name": "Schmoe",
                "line1": "234 5th Ave",
                "line4": "Manhattan",
                "postcode": "10001",
                "state": "NY",
                "country": reverse('country-detail', args=['US']),
                "phone_number": "+1 (717) 467-1111",
            },
            "payment": {
                "cybersource": {
                    "enabled": True,
                }
            }
        }
        data.update(extra_data)
        url = reverse('api-checkout')
        return self.client.post(url, data, format='json')


    def do_fetch_payment_states(self):
        return self.client.get(reverse('api-payment'))


    def do_cs_reply(self, cs_resp):
        soup = BeautifulSoup(cs_resp.content, 'html.parser')
        form_data = {}
        for element in soup.find_all('input'):
            form_data[element['name']] = element['value']
        # We have the data from cybersource, send it to our cybersource callback
        url = reverse('cybersource-reply')
        return self.client.post(url, form_data)


    def do_cs_get_token(self, cs_url, fields, extra_fields={}):
        next_year = datetime.date.today().year + 1
        cs_data = {
            'card_type': '001',
            'card_number': '4111111111111111',
            'card_cvn': '123',
            'card_expiry_date': '12-{}'.format(next_year),
        }
        for field in fields:
            if not field['editable'] or field['key'] not in cs_data:
                cs_data[field['key']] = field['value']
        cs_data.update(extra_fields)
        resp = requests.post(cs_url, cs_data)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        return self.do_cs_reply(resp)


    def do_cs_auth(self, cs_url, fields, extra_fields={}):
        cs_data = {}
        for field in fields:
            if not field['editable'] or field['key'] not in cs_data:
                cs_data[field['key']] = field['value']
        cs_data.update(extra_fields)
        resp = requests.post(cs_url, cs_data)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        return self.do_cs_reply(resp)


    def check_finished_order(self, number, product_id, quantity=1):
        # Order exists and was paid for
        self.assertEqual(Order.objects.all().count(), 1)
        order = Order.objects.get()
        self.assertEqual(order.number, number)

        lines = order.lines.all()
        self.assertEqual(lines.count(), 1)
        line = lines[0]
        self.assertEqual(line.quantity, quantity)
        self.assertEqual(line.product_id, product_id)

        payment_events = order.payment_events.filter(event_type__name="Authorise")
        self.assertEqual(payment_events.count(), 1)
        self.assertEqual(payment_events[0].amount, order.total_incl_tax)

        payment_sources = order.sources.all()
        self.assertEqual(payment_sources.count(), 1)
        self.assertEqual(payment_sources[0].currency, order.currency)
        self.assertEqual(payment_sources[0].amount_allocated, order.total_incl_tax)
        self.assertEqual(payment_sources[0].amount_debited, D('0.00'))
        self.assertEqual(payment_sources[0].amount_refunded, D('0.00'))

        transactions = payment_sources[0].transactions.all()
        self.assertEqual(transactions.count(), 1)
        self.assertEqual(transactions[0].txn_type, 'Authorise')
        self.assertEqual(transactions[0].amount, order.total_incl_tax)
        self.assertEqual(transactions[0].status, 'ACCEPT')

        self.assertEqual(transactions[0].log_field('req_reference_number'), order.number)
        self.assertEqual(transactions[0].token.card_last4, '1111')

        self.assertEqual(len(mail.outbox), 1)



class CheckoutIntegrationTest(BaseCheckoutTest):
    """Full Integration Test of Checkout"""

    def test_checkout_process(self):
        """Full checkout process using minimal api calls"""
        product = self.create_product()

        resp = self.do_get_basket()
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        basket_id = resp.data['id']

        resp = self.do_add_to_basket(product.id)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)

        resp = self.do_checkout(basket_id)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        order_number = resp.data['number']

        resp = self.do_fetch_payment_states()
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(resp.data['order_status'], 'Pending')
        self.assertEqual(resp.data['payment_method_states']['cybersource']['status'], 'Pending')
        self.assertEqual(resp.data['payment_method_states']['cybersource']['amount'], '10.00')
        self.assertEqual(resp.data['payment_method_states']['cybersource']['required_action']['name'], 'get-token')

        action = resp.data['payment_method_states']['cybersource']['required_action']
        resp = self.do_cs_get_token(action['url'], action['fields'])
        self.assertEqual(resp.status_code, status.HTTP_302_FOUND)

        resp = self.do_fetch_payment_states()
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(resp.data['order_status'], 'Pending')
        self.assertEqual(resp.data['payment_method_states']['cybersource']['status'], 'Pending')
        self.assertEqual(resp.data['payment_method_states']['cybersource']['amount'], '10.00')
        self.assertEqual(resp.data['payment_method_states']['cybersource']['required_action']['name'], 'authorize')

        action = resp.data['payment_method_states']['cybersource']['required_action']
        resp = self.do_cs_auth(action['url'], action['fields'])
        self.assertEqual(resp.status_code, status.HTTP_302_FOUND)

        resp = self.do_fetch_payment_states()
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(resp.data['order_status'], 'Authorized')
        self.assertEqual(resp.data['payment_method_states']['cybersource']['status'], 'Complete')
        self.assertEqual(resp.data['payment_method_states']['cybersource']['amount'], '10.00')
        self.assertIsNone(resp.data['payment_method_states']['cybersource']['required_action'])

        self.check_finished_order(order_number, product.id)


    def test_add_product_during_auth(self):
        """Test attempting to add a product during the authorize flow"""
        product = self.create_product()

        resp = self.do_get_basket()
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        basket_id = resp.data['id']

        # Adding a product here should succeed
        resp = self.do_add_to_basket(product.id)
        basket1 = resp.data['id']
        self.assertEqual(resp.status_code, status.HTTP_200_OK)

        resp = self.do_checkout(basket_id)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        order_number = resp.data['number']

        # Adding a product here should go to a new basket, not the one we're auth'ing
        resp = self.do_add_to_basket(product.id)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        basket2 = resp.data['id']
        self.assertNotEqual(basket1, basket2)

        # Finish checkout process
        resp = self.do_fetch_payment_states()
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(resp.data['order_status'], 'Pending')
        self.assertEqual(resp.data['payment_method_states']['cybersource']['required_action']['name'], 'get-token')

        action = resp.data['payment_method_states']['cybersource']['required_action']
        resp = self.do_cs_get_token(action['url'], action['fields'])
        self.assertEqual(resp.status_code, status.HTTP_302_FOUND)

        resp = self.do_fetch_payment_states()
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(resp.data['order_status'], 'Pending')
        self.assertEqual(resp.data['payment_method_states']['cybersource']['required_action']['name'], 'authorize')

        action = resp.data['payment_method_states']['cybersource']['required_action']
        resp = self.do_cs_auth(action['url'], action['fields'])
        self.assertEqual(resp.status_code, status.HTTP_302_FOUND)

        resp = self.do_fetch_payment_states()
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(resp.data['order_status'], 'Authorized')
        self.check_finished_order(order_number, product.id)

        # Adding a product here should go to basket2, not basket1
        resp = self.do_add_to_basket(product.id)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        basket3 = resp.data['id']
        self.assertEqual(basket2, basket3)


    def test_pay_for_nothing(self):
        """Test attempting to pay for an empty basket"""
        resp = self.do_get_basket()
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        basket_id = resp.data['id']

        resp = self.do_checkout(basket_id)
        self.assertEqual(resp.status_code, status.HTTP_406_NOT_ACCEPTABLE)


    def test_manipulate_total_pre_auth(self):
        """Test attempting to manipulate basket price when requesting an auth form"""
        product = self.create_product()

        resp = self.do_get_basket()
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        basket_id = resp.data['id']

        resp = self.do_add_to_basket(product.id)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(resp.data['total_incl_tax'], '10.00')

        resp = self.do_checkout(basket_id, extra_data={
            "total": "2.00",  # Try and get $10 of product for only $2
        })
        self.assertEqual(resp.status_code, status.HTTP_406_NOT_ACCEPTABLE)


    def test_manipulate_total_during_auth(self):
        """Test attempting to manipulate basket price when requesting auth from CyberSource"""
        product = self.create_product()

        resp = self.do_get_basket()
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        basket_id = resp.data['id']

        resp = self.do_add_to_basket(product.id)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(resp.data['total_incl_tax'], '10.00')

        resp = self.do_checkout(basket_id)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)

        resp = self.do_fetch_payment_states()
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(resp.data['order_status'], 'Pending')
        self.assertEqual(resp.data['payment_method_states']['cybersource']['status'], 'Pending')
        self.assertEqual(resp.data['payment_method_states']['cybersource']['amount'], '10.00')
        self.assertEqual(resp.data['payment_method_states']['cybersource']['required_action']['name'], 'get-token')

        action = resp.data['payment_method_states']['cybersource']['required_action']
        self.do_cs_get_token(action['url'], action['fields'])

        resp = self.do_fetch_payment_states()
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(resp.data['order_status'], 'Pending')
        self.assertEqual(resp.data['payment_method_states']['cybersource']['status'], 'Pending')
        self.assertEqual(resp.data['payment_method_states']['cybersource']['amount'], '10.00')
        self.assertEqual(resp.data['payment_method_states']['cybersource']['required_action']['name'], 'authorize')

        action = resp.data['payment_method_states']['cybersource']['required_action']
        next_year = datetime.date.today().year + 1
        cs_data = {
            'card_type': '001',
            'card_number': '4111111111111111',
            'card_cvn': '123',
            'card_expiry_date': '12-{}'.format(next_year),
            'bill_to_forename': 'Testy',
            'bill_to_surname': 'McUnitTest',
            'bill_to_address_line1': '234 5th Ave',
            'bill_to_address_line2': 'apt 5',
            'bill_to_address_city': 'Manhattan',
            'bill_to_address_state': 'NY',
            'bill_to_address_postal_code': '10001',
            'bill_to_address_country': 'US',
            'bill_to_phone': '17174671111',
        }
        for field in action['fields']:
            if not field['editable'] or field['key'] not in cs_data:
                cs_data[field['key']] = field['value']

        cs_data['amount'] = '2.00'  # Try and change the auth amount

        resp = requests.post(action['url'], cs_data)
        self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)


    def test_free_product(self):
        """Full checkout process using minimal api calls"""
        product = self.create_product(price=D('0.00'))

        resp = self.do_get_basket()
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        basket_id = resp.data['id']

        resp = self.do_add_to_basket(product.id)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)

        resp = self.do_checkout(basket_id)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        order_number = resp.data['number']

        resp = self.do_fetch_payment_states()
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(resp.data['order_status'], 'Pending')
        self.assertEqual(resp.data['payment_method_states']['cybersource']['required_action']['name'], 'get-token')

        action = resp.data['payment_method_states']['cybersource']['required_action']
        resp = self.do_cs_get_token(action['url'], action['fields'])
        self.assertEqual(resp.status_code, status.HTTP_302_FOUND)

        resp = self.do_fetch_payment_states()
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(resp.data['order_status'], 'Pending')
        self.assertEqual(resp.data['payment_method_states']['cybersource']['required_action']['name'], 'authorize')

        action = resp.data['payment_method_states']['cybersource']['required_action']
        resp = self.do_cs_auth(action['url'], action['fields'])
        self.assertEqual(resp.status_code, status.HTTP_302_FOUND)

        resp = self.do_fetch_payment_states()
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(resp.data['order_status'], 'Authorized')
        self.assertEqual(resp.data['payment_method_states']['cybersource']['amount'], '0.00')
        self.check_finished_order(order_number, product.id)



class CSReplyViewTest(BaseCheckoutTest):
    """Test the CybersourceReplyView with fixtured requests"""

    def prepare_order(self):
        """Setup a basket"""
        product = self.create_product()

        resp = self.do_get_basket()
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        basket_id = resp.data['id']

        resp = self.do_add_to_basket(product.id)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)

        resp = self.do_checkout(basket_id)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        order_number = resp.data['number']

        return order_number


    @patch('oscarapicheckout.signals.order_payment_authorized.send')
    def test_invalid_signature(self, order_payment_authorized):
        """Invalid signature should result in 400 Bad Request"""
        order_number = self.prepare_order()
        data = cs_factories.build_declined_token_reply_data(order_number)
        data = cs_factories.sign_reply_data(data)

        data['signature'] = 'abcdef'

        url = reverse('cybersource-reply')
        resp = self.client.post(url, data)
        self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(order_payment_authorized.call_count, 0, 'Should not trigger signal')
        self.assertEqual(self.do_fetch_payment_states().data['order_status'], 'Pending', 'Should not authorize')


    @patch('oscarapicheckout.signals.order_payment_authorized.send')
    def test_invalid_request_type(self, order_payment_authorized):
        """Bad request type should result in 400 Bad Request"""
        order_number = self.prepare_order()
        data = cs_factories.build_declined_token_reply_data(order_number)

        data["req_transaction_type"] = "payment",

        data = cs_factories.sign_reply_data(data)
        url = reverse('cybersource-reply')
        resp = self.client.post(url, data)
        self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(order_payment_authorized.call_count, 0, 'Should not trigger signal')
        self.assertEqual(self.do_fetch_payment_states().data['order_status'], 'Pending', 'Should not authorize')


    @patch('oscarapicheckout.signals.order_payment_authorized.send')
    def test_duplicate_transaction_id(self, order_payment_authorized):
        """Duplicate Transaction ID should result in redirect to the success page"""
        order_number = self.prepare_order()
        self.assertEqual(order_payment_authorized.call_count, 0)
        self.assertEqual(self.do_fetch_payment_states().data['order_status'], 'Pending', 'Should not authorize')

        data = cs_factories.build_accepted_token_reply_data(order_number)
        data = cs_factories.sign_reply_data(data)
        url = reverse('cybersource-reply')
        resp = self.client.post(url, data)

        self.assertRedirects(resp, reverse('checkout:index'), fetch_redirect_response=False)
        self.assertEqual(order_payment_authorized.call_count, 0)
        self.assertEqual(self.do_fetch_payment_states().data['order_status'], 'Pending')

        data = cs_factories.build_accepted_auth_reply_data(order_number)
        data = cs_factories.sign_reply_data(data)
        url = reverse('cybersource-reply')
        resp = self.client.post(url, data)

        self.assertRedirects(resp, reverse('checkout:thank-you'), fetch_redirect_response=False)
        self.assertEqual(order_payment_authorized.call_count, 1)
        self.assertEqual(self.do_fetch_payment_states().data['order_status'], 'Authorized')

        resp = self.client.post(url, data)
        self.assertRedirects(resp, reverse('checkout:thank-you'), fetch_redirect_response=False)
        self.assertEqual(order_payment_authorized.call_count, 1)
        self.assertEqual(self.do_fetch_payment_states().data['order_status'], 'Authorized')


    @patch('oscarapicheckout.signals.order_payment_authorized.send')
    def test_invalid_reference_number(self, order_payment_authorized):
        """Mismatched reference number should result in 400 Bad Request"""
        order_number = self.prepare_order()
        data = cs_factories.build_accepted_token_reply_data(order_number + 'ABC')
        data = cs_factories.sign_reply_data(data)
        url = reverse('cybersource-reply')
        resp = self.client.post(url, data)
        self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(order_payment_authorized.call_count, 0)
        self.assertEqual(self.do_fetch_payment_states().data['order_status'], 'Pending')


    @patch('oscarapicheckout.signals.order_payment_authorized.send')
    def test_declined_card(self, order_payment_authorized):
        """Declined card should should result in redirect to failure page"""
        order_number = self.prepare_order()
        data = cs_factories.build_declined_token_reply_data(order_number)
        data = cs_factories.sign_reply_data(data)
        url = reverse('cybersource-reply')

        resp = self.client.post(url, data)
        self.assertRedirects(resp, reverse('checkout:index'), fetch_redirect_response=False)

        self.assertEqual(order_payment_authorized.call_count, 0, 'Should not trigger signal')
        self.assertEqual(self.do_fetch_payment_states().data['order_status'], 'Payment Declined')


    @patch('oscarapicheckout.signals.order_payment_authorized.send')
    def test_success(self, order_payment_authorized):
        """Successful authorization should create an order and redirect to the success page"""
        order_number = self.prepare_order()

        data = cs_factories.build_accepted_token_reply_data(order_number)
        data = cs_factories.sign_reply_data(data)
        url = reverse('cybersource-reply')
        self.assertEqual(order_payment_authorized.call_count, 0)
        resp = self.client.post(url, data)

        data = cs_factories.build_accepted_auth_reply_data(order_number)
        data = cs_factories.sign_reply_data(data)
        url = reverse('cybersource-reply')
        self.assertEqual(order_payment_authorized.call_count, 0)
        resp = self.client.post(url, data)

        self.assertRedirects(resp, reverse('checkout:thank-you'))

        self.assertEqual(self.do_fetch_payment_states().data['order_status'], 'Authorized')
        self.assertEqual(order_payment_authorized.call_count, 1, 'Should trigger order_payment_authorized signal')

        order = order_payment_authorized.call_args[1]['order']
        self.assertEqual(order.status, 'Authorized', 'Should set order status')
        self.assertEqual(order.number, order_number, 'Should use order number from CS request')

        session = self.client.session
        self.assertEquals(session[CHECKOUT_ORDER_ID], order.id, 'Should save order_id in session')

        self.assertEqual(order.sources.count(), 1, 'Should save PaymentSource')
        source = order.sources.first()
        self.assertEqual(source.currency, 'USD')
        self.assertEqual(source.amount_allocated, D('99.99'))
        self.assertEqual(source.amount_refunded, D('0.00'))
        self.assertEqual(source.amount_debited, D('0.00'))

        self.assertEqual(source.transactions.count(), 1, 'Should save Transaction')
        transaction = source.transactions.first()
        self.assertEqual(transaction.token.masked_card_number, 'xxxxxxxxxxxx1111')
        self.assertEqual(transaction.token.card_type, '001')
        self.assertEqual(transaction.txn_type, 'Authorise')
        self.assertEqual(transaction.amount, D('99.99'))
        self.assertEqual(transaction.reference, data['transaction_id'])
        self.assertEqual(transaction.status, 'ACCEPT')
        self.assertEqual(transaction.request_token, data['request_token'])

        self.assertEqual(order.payment_events.count(), 1, 'Should save PaymentEvent')
        event = order.payment_events.first()
        self.assertEqual(event.amount, D('99.99'))
        self.assertEqual(event.reference, data['transaction_id'])
        self.assertEqual(event.event_type.name, 'Authorise')

        self.assertEqual(event.line_quantities.count(), 1, 'Should save PaymentEventQuantity')
        lq = event.line_quantities.first()
        self.assertEqual(lq.line, order.lines.first())
        self.assertEqual(lq.quantity, 1)


    def test_send_mail(self):
        """Successful authorization should send an email"""
        order_number = self.prepare_order()

        data = cs_factories.build_accepted_token_reply_data(order_number)
        data = cs_factories.sign_reply_data(data)
        url = reverse('cybersource-reply')
        resp = self.client.post(url, data)

        data = cs_factories.build_accepted_auth_reply_data(order_number)
        data = cs_factories.sign_reply_data(data)
        url = reverse('cybersource-reply')
        resp = self.client.post(url, data)

        self.assertRedirects(resp, reverse('checkout:thank-you'))

        self.assertEqual(self.do_fetch_payment_states().data['order_status'], 'Authorized')
        self.assertEqual(len(mail.outbox), 1, 'Should send email')



class CybersourceMethodTest(BaseCheckoutTest):
    @patch('cybersource.signals.pre_build_auth_request.send')
    @patch('cybersource.signals.pre_build_get_token_request.send')
    @patch('oscarapicheckout.signals.pre_calculate_total.send')
    def test_request_auth_form_success(self, pre_calculate_total, pre_build_get_token_request, pre_build_auth_request):
        product = self.create_product()

        resp = self.do_get_basket()
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        basket_id = resp.data['id']

        resp = self.do_add_to_basket(product.id)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)

        # Add some taxes to the basket
        def add_taxes(sender, basket, shipping_address, **kwargs):
            for line in basket.all_lines():
                line.purchase_info.price.tax = D('0.42')
        pre_calculate_total.side_effect = add_taxes

        # Add an extra field into the request
        def add_a_field(sender, extra_fields, **kwargs):
            extra_fields['my_custom_field'] = 'ABC'
        pre_build_get_token_request.side_effect = add_a_field
        pre_build_auth_request.side_effect = add_a_field

        # Checkout
        resp = self.do_checkout(basket_id)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        order_number = resp.data['number']
        self.assertEqual(pre_calculate_total.call_count, 1)

        # Fetch payment state
        resp = self.do_fetch_payment_states()
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(pre_build_get_token_request.call_count, 1)

        action = resp.data['payment_method_states']['cybersource']['required_action']
        cs_url = action['url']
        data = {}
        for field in action['fields']:
            data[field['key']] = field['value']

        # Check response fields
        self.assertEqual(cs_url, 'https://testsecureacceptance.cybersource.com/silent/pay')
        self.assertEquals(data['amount'], '10.42')
        self.assertEquals(data['bill_to_address_city'], 'Manhattan')
        self.assertEquals(data['bill_to_address_country'], 'US')
        self.assertEquals(data['bill_to_address_line1'], '234 5th Ave')
        self.assertEquals(data['bill_to_address_line2'], '')
        self.assertEquals(data['bill_to_address_postal_code'], '10001')
        self.assertEquals(data['bill_to_address_state'], 'NY')
        self.assertEquals(data['bill_to_email'], 'joe@example.com')
        self.assertEquals(data['bill_to_forename'], 'Joe')
        self.assertEquals(data['bill_to_phone'], '')
        self.assertEquals(data['bill_to_surname'], 'Schmoe')
        self.assertEquals(data['card_cvn'], '')
        self.assertEquals(data['card_expiry_date'], '')
        self.assertEquals(data['card_number'], '')
        self.assertEquals(data['card_type'], '')
        self.assertEquals(data['currency'], 'USD')
        self.assertEquals(data['customer_ip_address'], '127.0.0.1')
        self.assertEquals(data['device_fingerprint_id'], '')
        self.assertEquals(data['item_0_name'], 'My Product')
        self.assertEquals(data['item_0_quantity'], '1')
        self.assertEquals(data['item_0_unit_price'], '10.42')
        self.assertEquals(data['line_item_count'], '1')
        self.assertEquals(data['locale'], 'en')
        self.assertEquals(data['my_custom_field'], 'ABC')
        self.assertEquals(data['payment_method'], 'card')
        self.assertEquals(data['reference_number'], order_number)
        self.assertEquals(data['ship_to_address_city'], 'Manhattan')
        self.assertEquals(data['ship_to_address_country'], 'US')
        self.assertEquals(data['ship_to_address_line1'], '234 5th Ave')
        self.assertEquals(data['ship_to_address_line2'], '')
        self.assertEquals(data['ship_to_address_postal_code'], '10001')
        self.assertEquals(data['ship_to_address_state'], 'NY')
        self.assertEquals(data['ship_to_forename'], 'Joe')
        self.assertEquals(data['ship_to_phone'], '17174671111')
        self.assertEquals(data['ship_to_surname'], 'Schmoe')
        self.assertEquals(data['transaction_type'], 'create_payment_token')

        # Move onto step 2, authorize
        data['card_cvn'] = '123'
        data['card_expiry_date'] = '12-2020'
        data['card_number'] = '4111111111111111'
        data['card_type'] = '001'
        resp = requests.post(cs_url, data)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        resp = self.do_cs_reply(resp)
        self.assertEquals(resp.status_code, status.HTTP_302_FOUND)
        self.assertEqual(pre_build_auth_request.call_count, 1)

        # Fetch payment states again
        resp = self.do_fetch_payment_states()
        self.assertEqual(resp.status_code, status.HTTP_200_OK)

        action = resp.data['payment_method_states']['cybersource']['required_action']
        cs_url = action['url']
        data = {}
        for field in action['fields']:
            data[field['key']] = field['value']

        # Check response fields
        self.assertEqual(cs_url, 'https://testsecureacceptance.cybersource.com/silent/pay')
        self.assertEquals(data['amount'], '10.42')
        self.assertEquals(data['bill_to_address_city'], 'Manhattan')
        self.assertEquals(data['bill_to_address_country'], 'US')
        self.assertEquals(data['bill_to_address_line1'], '234 5th Ave')
        self.assertEquals(data['bill_to_address_line2'], '')
        self.assertEquals(data['bill_to_address_postal_code'], '10001')
        self.assertEquals(data['bill_to_address_state'], 'NY')
        self.assertEquals(data['bill_to_email'], 'joe@example.com')
        self.assertEquals(data['bill_to_forename'], 'Joe')
        self.assertEquals(data['bill_to_phone'], '')
        self.assertEquals(data['bill_to_surname'], 'Schmoe')
        self.assertEquals(data['currency'], 'USD')
        self.assertEquals(data['customer_ip_address'], '127.0.0.1')
        self.assertEquals(data['device_fingerprint_id'], '')
        self.assertEquals(data['item_0_name'], 'My Product')
        self.assertEquals(data['item_0_quantity'], '1')
        self.assertEquals(data['item_0_unit_price'], '10.42')
        self.assertEquals(data['line_item_count'], '1')
        self.assertEquals(data['locale'], 'en')
        self.assertEquals(data['my_custom_field'], 'ABC')
        self.assertEquals(data['payment_method'], 'card')
        self.assertEquals(data['reference_number'], order_number)
        self.assertEquals(data['ship_to_address_city'], 'Manhattan')
        self.assertEquals(data['ship_to_address_country'], 'US')
        self.assertEquals(data['ship_to_address_line1'], '234 5th Ave')
        self.assertEquals(data['ship_to_address_line2'], '')
        self.assertEquals(data['ship_to_address_postal_code'], '10001')
        self.assertEquals(data['ship_to_address_state'], 'NY')
        self.assertEquals(data['ship_to_forename'], 'Joe')
        self.assertEquals(data['ship_to_phone'], '17174671111')
        self.assertEquals(data['ship_to_surname'], 'Schmoe')
        self.assertEquals(data['transaction_type'], 'authorization')
        self.assertIsNotNone(data['payment_token'])



from django.conf import settings
from django.core.exceptions import ImproperlyConfigured


def overridable(name, default=None, required=False):
    if required:
        if not hasattr(settings, name) or not getattr(settings, name):
            raise ImproperlyConfigured("%s must be defined in Django settings" % name)
    return getattr(settings, name, default)


DEFAULT_CURRENCY = overridable('OSCAR_DEFAULT_CURRENCY', required=True)

PROFILE = overridable('CYBERSOURCE_PROFILE', required=True)
ACCESS = overridable('CYBERSOURCE_ACCESS', required=True)
SECRET = overridable('CYBERSOURCE_SECRET', required=True)

ORG_ID = overridable('CYBERSOURCE_ORG_ID', required=True)
MERCHANT_ID = overridable('CYBERSOURCE_MERCHANT_ID', required=True)

REDIRECT_PENDING = overridable('CYBERSOURCE_REDIRECT_PENDING', required=True)
REDIRECT_SUCCESS = overridable('CYBERSOURCE_REDIRECT_SUCCESS', required=True)
REDIRECT_FAIL = overridable('CYBERSOURCE_REDIRECT_FAIL', required=True)

ENDPOINT_PAY = overridable('CYBERSOURCE_ENDPOINT_PAY', 'https://testsecureacceptance.cybersource.com/silent/pay')

DATE_FORMAT = overridable('CYBERSOURCE_DATE_FORMAT', '%Y-%m-%dT%H:%M:%SZ')
LOCALE = overridable('CYBERSOURCE_LOCALE', 'en')

FINGERPRINT_PROTOCOL = overridable('CYBERSOURCE_FINGERPRINT_PROTOCOL', 'https')
FINGERPRINT_HOST = overridable('CYBERSOURCE_FINGERPRINT_HOST', 'h.online-metrix.net')

SOURCE_TYPE = overridable('CYBERSOURCE_SOURCE_TYPE', 'CyberSource Secure Acceptance')
CARD_REJECT_ERROR = overridable('CYBERSOURCE_CARD_REJECT_ERROR', 'Card was declined by the issuing bank. Please try a different card.')



CHECKOUT_ORDER_ID = 'checkout_order_id'
CHECKOUT_FINGERPRINT_SESSION_ID = 'cybersource_fingerprint_session_id'

# Default code for the email to send after successful checkout
DECISION_ACCEPT = 'ACCEPT'



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import django.contrib.postgres.fields.hstore
from django.contrib.postgres.operations import HStoreExtension
import django.db.models.deletion
import cybersource.models
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        HStoreExtension(),
        migrations.CreateModel(
            name='CyberSourceReply',
            fields=[
                ('id', models.AutoField(serialize=False, primary_key=True, verbose_name='ID', auto_created=True)),
                ('data', django.contrib.postgres.fields.hstore.HStoreField()),
                ('date_modified', models.DateTimeField(auto_now=True, verbose_name='Date Modified')),
                ('date_created', models.DateTimeField(auto_now_add=True, verbose_name='Date Received')),
                ('user', models.ForeignKey(blank=True, to=settings.AUTH_USER_MODEL, related_name='cybersource_replies', on_delete=django.db.models.deletion.SET_NULL, null=True)),  # NOQA
            ],
        ),
        migrations.CreateModel(
            name='PaymentToken',
            fields=[
                ('id', models.AutoField(serialize=False, primary_key=True, verbose_name='ID', auto_created=True)),
                ('token', models.CharField(unique=True, max_length=100)),
                ('masked_card_number', models.CharField(max_length=25)),
                ('card_type', models.CharField(max_length=10)),
                ('log', models.ForeignKey(related_name='tokens', to='cybersource.CyberSourceReply')),
            ],
            bases=(cybersource.models.ReplyLogMixin, models.Model),
        ),
    ]



#!/usr/bin/env python
import codecs
import os.path
from setuptools import setup
from versiontag import get_version, cache_git_tag


packages = [
    'cybersource',
    'cybersource.migrations',
    'cybersource.tests',
]

setup_requires = [
    'versiontag>=1.1.0',
]

requires = [
    'Django>=1.8.11',
    'django-oscar>=1.1.1',
    'django-oscar-api>=1.0.4',
    'django-oscar-api-checkout>=0.1.5',
]

def fpath(name):
    return os.path.join(os.path.dirname(__file__), name)

def read(fname):
    return codecs.open(fpath(fname), encoding='utf-8').read()

cache_git_tag()

setup(
    name='django-oscar-cybersource',
    description="Integration between django-oscar and the Cybersource Secure Acceptance.",
    version=get_version(pypi=True),
    long_description=open('README.rst').read(),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 1.8',
        'Framework :: Django :: 1.9',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: ISC License (ISCL)',
        'Operating System :: Unix',
        'Operating System :: MacOS :: MacOS X',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    author='Craig Weber',
    author_email='crgwbr@gmail.com',
    url='https://gitlab.com/thelabnyc/django-oscar-cybersource',
    license='ISC',
    packages=packages,
    install_requires=requires,
    setup_requires=setup_requires
)



from django.conf import settings
from django.core.urlresolvers import reverse
from django.test import TestCase


class FingerprintRedirectViewTest(TestCase):
    def test_img1(self):
        session = self.client.session
        session['cybersource_fingerprint_session_id'] = 'foo'
        session.save()
        url = reverse('cybersource-fingerprint-redirect', args=['img-1'])
        expected = 'https://h.online-metrix.net/fp/clear.png?org_id=%s&session_id=%s%s&m=1' % (
            settings.CYBERSOURCE_ORG_ID,
            settings.CYBERSOURCE_MERCHANT_ID,
            'foo')

        response = self.client.get(url)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, expected)

    def test_img2(self):
        session = self.client.session
        session['cybersource_fingerprint_session_id'] = 'bar'
        session.save()
        url = reverse('cybersource-fingerprint-redirect', args=['img-2'])
        expected = 'https://h.online-metrix.net/fp/clear.png?org_id=%s&session_id=%s%s&m=2' % (
            settings.CYBERSOURCE_ORG_ID,
            settings.CYBERSOURCE_MERCHANT_ID,
            'bar')

        response = self.client.get(url)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, expected)

    def test_flash(self):
        session = self.client.session
        session['cybersource_fingerprint_session_id'] = 'baz'
        session.save()
        url = reverse('cybersource-fingerprint-redirect', args=['flash'])
        expected = 'https://h.online-metrix.net/fp/fp.swf?org_id=%s&session_id=%s%s' % (
            settings.CYBERSOURCE_ORG_ID,
            settings.CYBERSOURCE_MERCHANT_ID,
            'baz')

        response = self.client.get(url)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, expected)

    def test_js(self):
        session = self.client.session
        session['cybersource_fingerprint_session_id'] = 'bat'
        session.save()
        url = reverse('cybersource-fingerprint-redirect', args=['js'])
        expected = 'https://h.online-metrix.net/fp/check.js?org_id=%s&session_id=%s%s' % (
            settings.CYBERSOURCE_ORG_ID,
            settings.CYBERSOURCE_MERCHANT_ID,
            'bat')

        response = self.client.get(url)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, expected)

    def test_unknown(self):
        url = reverse('cybersource-fingerprint-redirect', args=['something'])
        response = self.client.get(url)
        self.assertEqual(response.status_code, 404)
